#include "CRecorder.h"

#include "GameIncludes.h"

CRecorder::CRecorder(void) :
_recordInterval(0.1f),
_recordRemainingTime(_recordInterval)
{
	
}


CRecorder::~CRecorder(void)
{
}

void CRecorder::addRecordingObject( CNode* RecordingObject )
{
	if (std::find(_recordingObjects.begin(), _recordingObjects.end(), RecordingObject) == _recordingObjects.end())
	{
		_recordingObjects.push_back(RecordingObject);
		CRecord record;
		record.saveFrom(RecordingObject);
		_recordingHistory[RecordingObject].addRecord(CGame::getInstance()->getGameTime(), record);
	}
}

void CRecorder::removeRecordingObject( CNode* RecordingObject )
{
	auto ItFound = std::find(_recordingObjects.begin(), _recordingObjects.end(), RecordingObject);
	if (ItFound != _recordingObjects.end())
	{
		_recordingObjects.erase(ItFound);
		_recordingHistory.erase(RecordingObject);
	}
}

void CRecorder::update( float DeltaTime )
{
	_recordRemainingTime -= DeltaTime;
	if (_recordRemainingTime <= 0.0f && DeltaTime >= 0.0f)
	{
		_recordRemainingTime += _recordInterval;
		this->recordObjects();
	}
	else if (DeltaTime <= 0.0f)
	{
		if (_recordRemainingTime >= _recordInterval)
		{
			_recordRemainingTime -= _recordInterval;
		}
		this->unRecordObjects();
	}

	this->manageActionRecords(DeltaTime);

	this->replayObjects(DeltaTime);

	//update current time of replaying objects
	for (auto ItObject = _replayingObjects.begin(); ItObject != _replayingObjects.end(); ++ItObject)
	{
		ItObject->second._time += DeltaTime;
	}
}

void CRecorder::recordObjects()
{
	for (auto ItObject = _recordingObjects.begin(); ItObject != _recordingObjects.end(); ++ItObject)
	{
		auto Object = (*ItObject);

		CRecord Record;
		Record.saveFrom(Object);
		_recordingHistory[Object].addRecord(CGame::getInstance()->getGameTime(), Record);
	}
}

void CRecorder::recordActionForObject( CNode* Object, EAction action )
{
	_recordingHistory[Object].addActionRecord(CGame::getInstance()->getGameTime(), action);
}

void CRecorder::clearAll()
{
	_recordingHistory.clear();
	_recordingObjects.clear();
	_recordRemainingTime = 0.0f;
	_replayingObjects.clear();
}

bool CRecorder::isRecordingObject( CNode* RecordingObject )
{
	auto ItFound = std::find(_recordingObjects.begin(), _recordingObjects.end(), RecordingObject);
	return ItFound != _recordingObjects.end();
}

CRecorder* CRecorder::getInstance()
{
	static CRecorder instance;
	return &instance;
}

void CRecorder::addReplayingObject( CNode* ReplayingObject, CNode* BasedOnRecordedObject, float Time )
{
	_replayingObjects[ReplayingObject]._basedOn = BasedOnRecordedObject;
	_replayingObjects[ReplayingObject]._time = Time;

	auto itBasedOn = _recordingHistory.find(BasedOnRecordedObject);
	if (itBasedOn != _recordingHistory.end())
	{
		itBasedOn->second.loadFromRecord(Time, ReplayingObject);
	}
}

void CRecorder::removeReplayingObject( CNode* ReplayingObject )
{
	_replayingObjects.erase(ReplayingObject);
}

void CRecorder::replayObjects(float DeltaTime)
{
	for (auto ItObject = _replayingObjects.begin(); ItObject != _replayingObjects.end(); ++ItObject)
	{
		if (ItObject->first->_bDeleted)
			continue;

		auto itBasedOn = _recordingHistory.find(ItObject->second._basedOn);
		if (itBasedOn != _recordingHistory.end())
		{
			itBasedOn->second.loadFromRecord(ItObject->second._time, ItObject->first);
		}

		//manage actions
		if (DeltaTime >= 0.f)
		{
			float OldTime = ItObject->second._time - DeltaTime;
			_recordingHistory[ItObject->second._basedOn].replayActionRecordFor(ItObject->first, OldTime, OldTime + DeltaTime);
		}
	}
}

bool CRecorder::isReplayingObject( CNode* Node )
{
	auto ItFound = _replayingObjects.find(Node);
	return ItFound != _replayingObjects.end();
}

void CRecorder::unRecordObjects()
{
	for (auto ItObject = _recordingObjects.begin(); ItObject != _recordingObjects.end(); ++ItObject)
	{
		auto Object = (*ItObject);

		_recordingHistory[Object].removePassedRecords(CGame::getInstance()->getGameTime());
	}
}

CNode* CRecorder::findMoreRecentReplayingObjectsBasedOn( CNode* BasedOnNode )
{
	CNode* found = nullptr;
	float time = FLT_MAX;
	for (auto ItObject = _replayingObjects.begin(); ItObject != _replayingObjects.end(); ++ItObject)
	{
		if (ItObject->second._basedOn == BasedOnNode && ItObject->second._time < time)
		{
			time = ItObject->second._time;
			found = ItObject->first;
		}
	}
	return found;
}

void CRecorder::replayRecordedObject( CNode* RecordedObject )
{
	auto ItFound = _recordingHistory.find(RecordedObject);
	if (ItFound != _recordingHistory.end())
	{
		ItFound->second.loadFromRecord(CGame::getInstance()->getGameTime(), ItFound->first);
	}
}

void CRecorder::manageActionRecords( float DeltaTime )
{
	if (DeltaTime <= 0.0f)
	{
		float OldTime = CGame::getInstance()->getGameTime() - DeltaTime;

		for (auto ItObject = _recordingObjects.begin(); ItObject != _recordingObjects.end(); ++ItObject)
		{
			auto Object = (*ItObject);

			_recordingHistory[Object].removeActionRecordIfNeeded(Object, OldTime, OldTime + DeltaTime);
		}
	}
}

void CRecorder::makeRecordIfRegistered( CNode* Node )
{
	auto ItFound = std::find(_recordingObjects.begin(), _recordingObjects.end(), Node);
	if (ItFound != _recordingObjects.end())
	{
		CRecord Record;
		Record.saveFrom(Node);
		_recordingHistory[Node].addRecord(CGame::getInstance()->getGameTime(), Record);
	}
}

bool CRecorder::isObjectBasedOnPlayedAround( CNode* BasedOnObject, float time, float marginTime )
{
	for (auto itObject = _replayingObjects.begin(); itObject != _replayingObjects.end(); ++itObject)
	{
		if (itObject->second._time <= time + marginTime && itObject->second._time >= time - marginTime && !itObject->first->_bDeleted)
			return true;
	}
	return false;
}

void CRecord::saveFrom( CNode* object )
{
	sf::Transformable* transformable = object->getTransformable();
	if (transformable)
	{
		_position = transformable->getPosition();
		_rotation = transformable->getRotation();
		_origin = transformable->getOrigin();
		_scale = transformable->getScale();
	}

	_health = object->getHealth();
	_velocity = object->_velocity;
}

void CRecord::loadTo( CNode* object )
{
	sf::Transformable* transformable = object->getTransformable();
	if (transformable)
	{
		transformable->setPosition(_position);
		transformable->setOrigin(_origin);
		transformable->setRotation(_rotation);
		transformable->setScale(_scale);
	}

	object->setHealth(_health);
	object->_velocity = _velocity;
}

void CRecord::loadFromInterpolationTo( CNode* object, const CRecord& Before, const CRecord& After, float Coeff )
{
	//interpolate between two values
	CRecord NewRecord;
	NewRecord._position = CMath::Blend(Before._position, After._position, Coeff);
	NewRecord._origin = CMath::Blend(Before._origin, After._origin, Coeff);
	NewRecord._rotation = CMath::Blend(Before._rotation, After._rotation, Coeff);
	NewRecord._scale = CMath::Blend(Before._scale, After._scale, Coeff);
	NewRecord._health = CMath::Blend(Before._health, After._health, Coeff);
	NewRecord._velocity = CMath::Blend(Before._velocity, After._velocity, Coeff);

	NewRecord.loadTo(object);
}

void CRecorderHistory::addRecord( float time, const CRecord& Record )
{
	_records.insert(std::make_pair(time, Record));
	_lastRecordSaved = std::make_pair(time, Record);
}

void CRecorderHistory::removePassedRecords(float time)
{
	if (_records.empty())
		return;

	auto itUp = _records.upper_bound(time);
	if (itUp != _records.end())
	{
		//save it 
		_lastRecordSaved = *itUp;
	}
	for (auto ItRecord = itUp; ItRecord != _records.end(); )
	{
		//erase it
		ItRecord = _records.erase(ItRecord); //absolutely
	}
}

void CRecorderHistory::loadFromRecord( float time, CNode* Object )
{
	//find two records
	auto ItBefore = _records.lower_bound(time);
	if (ItBefore != _records.begin())
	{
		--ItBefore;
	}
	auto ItAfter = _records.upper_bound(time);
	if (ItBefore != _records.end() && ItAfter != _records.end() && ItBefore != ItAfter)
	{
		float Coeff = (time - (*ItBefore).first) / ((*ItAfter).first - (*ItBefore).first);

		CRecord::loadFromInterpolationTo(Object, ItBefore->second, ItAfter->second, Coeff);
	}
	else if (ItBefore != _records.end())
	{
		if (_lastRecordSaved.first < time)
		{
			CRecord record;
			record.saveFrom(Object);
			_lastRecordSaved = std::make_pair(time, record);
		}
		//interpolate between two values
		float Coeff = (time - (*ItBefore).first) / (_lastRecordSaved.first - (*ItBefore).first);

		CRecord::loadFromInterpolationTo(Object, ItBefore->second, _lastRecordSaved.second, Coeff);
	}
	else if (ItAfter != _records.end())
	{
		(*ItAfter).second.loadTo(Object);
	}
}

void CRecorderHistory::replayActionRecordFor( CNode* Object, float OldTime, float NewTime ) const
{
	auto ItLB = _actionRecords.lower_bound(OldTime);
	auto ItUB = _actionRecords.upper_bound(NewTime);
	for (auto ItAction = ItLB; ItAction != ItUB; ++ItAction)
	{
		Object->doAction(ItAction->second);
	}
}

void CRecorderHistory::addActionRecord( float time, EAction action )
{
	_actionRecords.insert(std::make_pair(time, action));
}

void CRecorderHistory::removeActionRecordIfNeeded( CNode* Object, float OldTime, float NewTime )
{
	if (NewTime < OldTime)
	{
		std::swap(OldTime, NewTime);
	}
	auto ItLB = _actionRecords.lower_bound(OldTime);
	auto ItUB = _actionRecords.upper_bound(NewTime);
	for (auto ItAction = ItLB; ItAction != ItUB && ItAction != _actionRecords.end(); )
	{
		ItAction = _actionRecords.erase(ItAction);
	}
}

