#include "CSerializer.h"

#include "GameIncludes.h"
#include "tinyxml2.h"

const std::string CSerializer::LevelSceneName = "scene";
const std::string CSerializer::GroupName = "group";
const std::string CSerializer::SpriteName = "sprite";
const std::string CSerializer::TextName = "text";


CSerializer::CSerializer(void)
{
	_defaultFont = new sf::Font();
	_defaultFont->loadFromFile("media/fonts/arial.ttf");
}


CSerializer::~CSerializer(void)
{
	delete _defaultFont;
}

CNode* CSerializer::loadFromXml( const tinyxml2::XMLElement& Element )
{
	CNode* node = NULL;
	if (Element.Name() == SpriteName)
	{
		sf::Sprite* sprite = this->createSprite(Element);
		node = new CNode();
		node->setDrawable(sprite);
		parseTransformableElement(Element, sprite);
	}
	else if (Element.Name() == TextName)
	{
		sf::Text* text = this->createText(Element);
		node = new CNode();
		node->setDrawable(text);
		parseTransformableElement(Element, text);
	}
	else 
	{
		CGroup* group = new CGroup();
		for (auto Child = Element.FirstChildElement(); Child; Child = Child->NextSiblingElement())
		{
			CNode* Node = loadFromXml(*Child);
			if (Node)
			{
				group->_children.push_back(Node);
			}
		}
		node = group;
	}

	tinyxml2::XMLPrinter printer;
	Element.Accept(&printer);
	node->setOriginalXml(printer.CStr());
	node->setDamage(Element.FloatAttribute("damage"));
	node->setScore(Element.IntAttribute("score"));

	addNodeBehaviors(Element, node);

	return node;
}



CLevel* CSerializer::loadLevelFromXmlFile( const std::string& path )
{
	tinyxml2::XMLDocument XmlDoc;

	if (XmlDoc.LoadFile(path.c_str()) != tinyxml2::XML_SUCCESS)
	{
		throw std::exception((std::string("Error loading level ") + path).c_str());
	}
	else
	{
		CLevel* level = new CLevel();

		const tinyxml2::XMLElement* LevelRootElement = FindDeepFirstChildElement(XmlDoc.FirstChildElement(), LevelSceneName);
		if (LevelRootElement)
		{
			level->_rootGroup = dynamic_cast<CGroup*>(loadFromXml(*LevelRootElement));
		}

		//load difficulty settings
		auto DiffElement = FindDeepFirstChildElement(XmlDoc.FirstChildElement(), "difficulty");
		if (DiffElement)
		{
			float time = 0.0f;
			for (auto Child = DiffElement->FirstChildElement(); Child; Child = Child->NextSiblingElement())
			{
				// <level copyinterval="5" fireinterval="1" firespeed="500" movespeed="300" rotationspeed="200"/>
				CDifficultyParams DiffParams;
				DiffParams._copyInterval = Child->FloatAttribute("copyinterval");
				DiffParams._fireInterval = Child->FloatAttribute("fireinterval");
				DiffParams._fireSpeed = Child->FloatAttribute("firespeed");
				DiffParams._moveSpeed = Child->FloatAttribute("movespeed");
				DiffParams._rotationSpeed = Child->FloatAttribute("rotationspeed");

				level->_difficultyParams.insert(std::make_pair(time, DiffParams));
				time += 10.0f; // 10 seconds theme hourra !
			}
		}

		return level;
	}
}

CSerializer* CSerializer::getInstance()
{
	static CSerializer instance;
	return &instance;
}

void CSerializer::parseTransformableElement( const tinyxml2::XMLElement& Element, sf::Transformable* Transformable )
{
	Transformable->setPosition(Element.FloatAttribute("x"), Element.FloatAttribute("y"));
	Transformable->setRotation(Element.FloatAttribute("rot"));
	Vector2f Scale = Transformable->getScale();
	Element.QueryFloatAttribute("xscale", &Scale.x);
	Element.QueryFloatAttribute("yscale", &Scale.y);
	Transformable->setScale(Scale);
}

sf::Text* CSerializer::createText( const tinyxml2::XMLElement& Element )
{
	sf::Text* text = new sf::Text();
	text->setString(Element.Attribute("string"));
	text->setFont(*_defaultFont);
	unsigned int charSize = 50;
	Element.QueryUnsignedAttribute("size", &charSize);
	text->setCharacterSize(charSize);
	
	unsigned int r, g, b, a;
	r = g = b = a = 255;
	Element.QueryUnsignedAttribute("red", &r);
	Element.QueryUnsignedAttribute("green", &g);
	Element.QueryUnsignedAttribute("blue", &b);
	Element.QueryUnsignedAttribute("alpha", &a);
	sf::Color color(r,g,b,a);

	text->setColor(color);

	return text;
}

sf::Sprite* CSerializer::createSprite( const tinyxml2::XMLElement& Element )
{
	sf::Texture* texture = new sf::Texture();
	texture->loadFromFile(Element.Attribute("path"));
	sf::Sprite* sprite = new sf::Sprite(*texture);
	sprite->setOrigin(texture->getSize().x / 2.0f, texture->getSize().y / 2.0f);
	return sprite;
}

const tinyxml2::XMLElement* CSerializer::FindDeepFirstChildElement( const tinyxml2::XMLElement* Parent, const std::string& name )
{
	for (auto Child = Parent->FirstChildElement(); Child; Child = Child->NextSiblingElement())
	{
		if (Child->Name() == name)
		{
			return Child;
		}
		auto returnChild = FindDeepFirstChildElement(Child, name);
		if (returnChild)
		{
			return returnChild;
		}
	}
	return nullptr;
}

void CSerializer::addNodeBehaviors( const tinyxml2::XMLElement& Element, CNode* node )
{
	if (Element.FindAttribute("PlayerController"))
	{
		node->addNodeBehavior(new CPlayerControllerBH());
	}
	else if (Element.FindAttribute("Score"))
	{
		node->addNodeBehavior(new CScoreBehavior());
	}
	else if (Element.FindAttribute("Timer"))
	{
		node->addNodeBehavior(new CTimerBehavior());
	}
	else if (Element.FindAttribute("Moving"))
	{
		CMovingBehavior* MovingBH = new CMovingBehavior();
		MovingBH->_direction.x = Element.FloatAttribute("dirx");
		MovingBH->_direction.y = Element.FloatAttribute("diry");
		MovingBH->_speed = Element.FloatAttribute("speed");
		node->addNodeBehavior(MovingBH);
	}
}

CNode* CSerializer::getCopyOf( const CNode* originalNode )
{
	tinyxml2::XMLDocument XmlDoc;
	if (XmlDoc.Parse(originalNode->getOriginalXml().c_str()) == tinyxml2::XML_NO_ERROR)
	{
		return this->loadFromXml(*XmlDoc.FirstChildElement());
	}
	return nullptr;
}

CNode* CSerializer::loadFromXmlFile( const std::string& file )
{
	tinyxml2::XMLDocument XmlDoc;
	XmlDoc.LoadFile(file.c_str());
	return loadFromXml(*XmlDoc.FirstChildElement());
}
